# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Dashboard": "core.dashboards.ipynb",
         "BaseGallery": "core.dashboards.ipynb",
         "RecordDastasetGallery": "core.dashboards.ipynb",
         "DatasetOverview": "core.dashboards.ipynb",
         "MultiDatasetOverview": "core.dashboards.ipynb",
         "DatasetComparison": "core.dashboards.ipynb",
         "DatasetFilter": "core.dashboards.ipynb",
         "DatasetFilterWithRangeSliderAndMultiSelect": "core.dashboards.ipynb",
         "DatasetFilterWithScatter": "core.dashboards.ipynb",
         "DatasetGenerator": "core.dashboards.ipynb",
         "DatasetGeneratorScatter": "core.dashboards.ipynb",
         "Observable": "core.data.ipynb",
         "ObservableList": "core.data.ipynb",
         "DatasetDescriptor": "core.data.ipynb",
         "StringDescriptor": "core.data.ipynb",
         "GenericDataset": "core.data.ipynb",
         "ObjectDetectionDatasetOverview": "dashboards.ipynb",
         "ObjectDetectionDatasetComparison": "dashboards.ipynb",
         "ObjectDetectionDatasetGeneratorScatter": "dashboards.ipynb",
         "ObjectDetectionDatasetGeneratorRange": "dashboards.ipynb",
         "ObjectDetectionResultOverview": "dashboards.ipynb",
         "InstanceSegmentationDatasetOverview": "dashboards.ipynb",
         "InstanceSegmentationDatasetComparison": "dashboards.ipynb",
         "InstanceSegmentationDatasetGeneratorScatter": "dashboards.ipynb",
         "InstanceSegmentationDatasetGeneratorRange": "dashboards.ipynb",
         "InstanceSegmentationResultOverview": "dashboards.ipynb",
         "RecordDataframeParser": "data.ipynb",
         "RecordDataset": "data.ipynb",
         "ResultsDataset": "data.ipynb",
         "BboxRecordDataframeParser": "data.ipynb",
         "DataDescriptorBbox": "data.ipynb",
         "StatsDescriptorBbox": "data.ipynb",
         "ImageStatsDescriptorBbox": "data.ipynb",
         "ClassStatsDescriptorBbox": "data.ipynb",
         "GalleryStatsDescriptorBbox": "data.ipynb",
         "BboxRecordDataset": "data.ipynb",
         "PrecisionRecallMetricsDescriptorObjectDetection": "data.ipynb",
         "ObjectDetectionResultsDataset": "data.ipynb",
         "InstanceSegmentationRecordDataframeParser": "data.ipynb",
         "DataDescriptorInstanceSegmentation": "data.ipynb",
         "StatsDescriptorInstanceSegmentation": "data.ipynb",
         "ImageStatsDescriptorInstanceSegmentation": "data.ipynb",
         "ClassStatsDescriptorInstanceSegmentation": "data.ipynb",
         "GalleryStatsDescriptorInstanceSegmentation": "data.ipynb",
         "InstanceSegmentationRecordDataset": "data.ipynb",
         "PrecisionRecallMetricsDescriptorInstanceSegmentation": "data.ipynb",
         "InstanceSegmentationResultsDataset": "data.ipynb",
         "APObjectDetection": "metrics.ipynb",
         "APInstanceSegmentation": "metrics.ipynb",
         "Filter": "plotting.controls.ipynb",
         "RangeFilter": "plotting.controls.ipynb",
         "CategoricalFilter": "plotting.controls.ipynb",
         "TimeFilter": "plotting.controls.ipynb",
         "ScatterFilter": "plotting.controls.ipynb",
         "GenericMulitScatterFilter": "plotting.controls.ipynb",
         "barplot": "plotting.core.ipynb",
         "histogram": "plotting.core.ipynb",
         "heatmap": "plotting.core.ipynb",
         "time_arc_plot": "plotting.core.ipynb",
         "table_from_dataframe": "plotting.core.ipynb",
         "stacked_hist": "plotting.core.ipynb",
         "categorical_2d_histogram": "plotting.core.ipynb",
         "categorical_2d_histogram_with_gui": "plotting.core.ipynb",
         "scatter_plot_with_gui": "plotting.core.ipynb",
         "plots_as_matrix": "plotting.core.ipynb",
         "toggle_legend_js": "plotting.utils.ipynb",
         "calculate_mixing_matrix": "plotting.utils.ipynb",
         "get_min_and_max_dates": "plotting.utils.ipynb",
         "convert_rgb_image_to_bokeh_rgb_image": "plotting.utils.ipynb",
         "draw_record_with_bokeh": "plotting.utils.ipynb",
         "erles_to_string": "utils.ipynb",
         "erles_to_counts_to_utf8": "utils.ipynb",
         "string_to_erles": "utils.ipynb",
         "correct_mask": "utils.ipynb",
         "decorrect_mask": "utils.ipynb"}

modules = ["core/dashboards.py",
           "core/data.py",
           "dashboards.py",
           "data.py",
           "metrics.py",
           "plotting/controls.py",
           "plotting/core.py",
           "plotting/utils.py",
           "utils.py"]

doc_url = "https://fstroth.github.io/icevision_dashboards/"

git_url = "https://github.com/fstroth/icevision_dashboards/tree/master/"

def custom_doc_links(name): return None
